from torch import nn
from torch.nn import functional as F
import torch
import numpy as np

class PositionalEncoding(nn.Module):
    def __init__(self, d_model, max_len=1000):
        super().__init__()
        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len).unsqueeze(1).float()
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-np.log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        self.register_buffer("pe", pe.unsqueeze(0))

    def forward(self, x):
        return x + self.pe[:, :x.size(1)]

class ContextEncoder(nn.Module):
    def __init__(self, input_dim, d_model=256, nhead=8, num_layers=4, ff_dim=512):
        super().__init__()
        self.input_proj = nn.Linear(input_dim, d_model)
        self.pos_enc = PositionalEncoding(d_model)
        encoder_layer = nn.TransformerEncoderLayer(
            d_model=d_model, nhead=nhead,
            dim_feedforward=ff_dim, dropout=0.1, batch_first=True
        )
        self.encoder = nn.TransformerEncoder(encoder_layer, num_layers)

    def forward(self, x):
        x = self.input_proj(x)
        x = self.pos_enc(x)
        return self.encoder(x)


import editdistance
import numpy as np
def metric(gt,pred):
    correct_chars = len(gt) - editdistance.eval(pred, gt)
    return correct_chars / len(gt)

def general_car(all_ref,all_hyp):
    all_scores = []
    for ref,hyp in zip(all_ref,all_hyp):
        all_scores.append(metric(ref,hyp))
    return np.mean(all_scores)

def general_war(all_ref,all_hyp):
    return metric(all_ref,all_hyp)


from torch import nn
import torch
import numpy as np

def encode_label(label_str, token2id):
    tokens = label_str.split('|')
    ids = []
    for t in tokens:
        if t in token2id:
            ids.append(token2id[t])
    return ids

class DecoderCTC(nn.Module):
    def __init__(self, d_model, num_classes):
        super().__init__()
        self.linear = nn.Linear(d_model, num_classes)

    def forward(self, x):
        return self.linear(x)

def ImgPad(image):
    image = np.array(image)
    image = np.expand_dims(np.expand_dims(image,axis=0),axis=0)
    image = torch.tensor(image,dtype=torch.float32)
    image.shape
    return image


import torch
import torch.nn as nn
import torchvision.models as models

class AryadCNN(nn.Module):
    def __init__(self):
        super().__init__()
        resnet = models.resnet50(weights=models.ResNet50_Weights.IMAGENET1K_V1)
        self.cnn = nn.Sequential(*list(resnet.children())[:-2])
        # Adapter pour images 1 canal (grayscale)
        original_conv1 = self.cnn[0]
        self.cnn[0] = nn.Conv2d(1, original_conv1.out_channels,
                                kernel_size=original_conv1.kernel_size,
                                stride=original_conv1.stride,
                                padding=original_conv1.padding,
                                bias=False)
        with torch.no_grad():
            self.cnn[0].weight.copy_(original_conv1.weight.sum(dim=1, keepdim=True))

    def forward(self, x):
        features = self.cnn(x)  # (B, C, H', W')
        B, C, H, W = features.shape
        features = features.permute(0, 3, 1, 2)  # (B, W', C, H')
        features = features.contiguous().view(B, W, -1)  # (B, W', C*H')
        return features  # (batch_size, seq_len, feature_dim)


import pandas as pd
from tqdm import tqdm
from glob import glob
import cv2
import re
from datasets import Dataset, Features, Array2D,Value,load_dataset


def processing_phoneme(dirs_files = {
    'set_a':"C:/Users/farya/Documents/OcrDataEtude/set_a/set_a/tru/*.tru",
    'set_b':"C:/Users/farya/Documents/OcrDataEtude/set_b/set_b/tru/*.tru",
    'set_c':"C:/Users/farya/Documents/OcrDataEtude/set_c/set_c/tru/*.tru",
    'set_d':"C:/Users/farya/Documents/OcrDataEtude/set_d/set_d/tru/*.tru",
},path_dataset="aryadacademie/Arabic_Ocr_set"):
    
    for name_file ,dir_file in tqdm(dirs_files.items()):
        try:
            tru = glob(dir_file)
            ids_files = []
            phonemes_files = []
            for path in tru:
                with open(f"{path}","r") as f:
                    verite = f.readlines()
                    id_file = re.search(r"COM:\s+(\S+\.tif)", verite[3]).group(1)[:-4]
                    phoneme_file = re.search(r"AW2:([^;]+)", verite[6]).group(1).rstrip('|')
                    ids_files.append(id_file)
                    phonemes_files.append(phoneme_file)
    
            df = pd.DataFrame({
                "id": ids_files,
                "phonemes": phonemes_files
            })
    
            features = Features({
                "id": Value("string"),
                "phonemes": Value("string")
            })
    
            dataset = Dataset.from_dict(df, features=features)
            dataset.push_to_hub(path_dataset, split=f"arabic_ocr_phonemes_{name_file}")
            print(f"Dataset {name_file} push avec succès\n")
    
        except Exception as e:
            print(f"Erreur avec {name_file} : {e}\n")


def processing_array(dir_image_resize = {
    "array_set_a":'C:/Users/farya/Documents/OcrDataEtude/set_a/set_a/tif/*.tif',
    "array_set_b":'C:/Users/farya/Documents/OcrDataEtude/set_b/set_b/tif/*.tif',
    "array_set_c":'C:/Users/farya/Documents/OcrDataEtude/set_c/set_c/tif/*.tif',
    "array_set_d":'C:/Users/farya/Documents/OcrDataEtude/set_d/set_d/tif/*.tif'
},path_dataset="aryadacademie/arabic_ocr_array_set_all"):
    for name_dir,path_dir in tqdm(dir_image_resize.items()):
        try:
            ids_paths = []
            imgs_paths = []
            images_paths = {re.search(r'([^\\/]+)\.tif$',path).group(1): cv2.resize(cv2.bitwise_not(cv2.imread(path.replace('\\','/'),0)),(300,100))//255  for path in glob(path_dir)}
            for id_path,img_path in images_paths.items():
                ids_paths.append(id_path)
                imgs_paths.append(img_path)
            
            ds = pd.DataFrame({
            "id":ids_paths,
            "array":imgs_paths}) 
            
            features = Features({
            "id": Value("string"),
            "array": Array2D(dtype="uint8", shape=(100,300))}) 
            
            dataset = Dataset.from_dict(ds, features=features)
            dataset.push_to_hub(path_dataset, split=f"arabic_ocr_{name_dir}")
            print(f"dataset {name_dir} push avec succès")
        
        except Exception as e:
            print(f"Erreur avec {name_dir} : {e}\n")


def merge_set(liste_set=['a','b','c','d'],
              ds_array = load_dataset("aryadacademie/arabic_ocr_array_set_all"),
              ds_ph = load_dataset("aryadacademie/Arabic_Ocr_set")
             ):
    for lettre in liste_set:
        ids_array = ds_array[f'arabic_ocr_array_set_{lettre}']['id']
        ids_ph = ds_ph[f'arabic_ocr_phonemes_set_{lettre}']['id']
        
        assert all(i == j for i, j in zip(ids_array, ids_ph))
        
        merged_dataset_set = ds_array[f'arabic_ocr_array_set_{lettre}'].add_column(
            'phonemes', 
            ds_ph[f'arabic_ocr_phonemes_set_{lettre}']['phonemes']
        )
        merged_dataset_set.push_to_hub("aryadacademie/Ocr_Arabic_Merged_data_set", split=f"Ocr_Arabic_Merged_data_set_{lettre}")
        print(f"dataset set_{lettre} push avec succès")


def vocab_processing(vocabs):
    for name_set,vocab in vocabs.items():
        vocab_set_token = set([token for phoneme in vocab for token in phoneme.split('|')])
        ds = {
            f'vocabulaire_{name_set}':list(vocab_set_token)
        }
        df = pd.DataFrame(ds)
        df.to_csv(f"vocabulaire_{name_set}.csv", index=False)

def read_csv_vocab(path):
    df = pd.read_csv(path)
    vocabOcrArabic = {token:id_ for token,id_ in zip(df['token'],df['id'])}
    return  vocabOcrArabic 


from .aryadcnn import AryadCNN
import torch
import torch.nn as nn
from .context_encoder import ContextEncoder
from .utils import DecoderCTC
from .metrics import general_car,general_war
import numpy as np
import os
from tqdm import tqdm


os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3' 
os.environ['XLA_FLAGS'] = '--xla_gpu_cuda_data_dir=/usr/local/cuda'
import tensorflow.compat.v1 as tf

tf.disable_v2_behavior()
tf.logging.set_verbosity(tf.logging.ERROR)




class AryadOcr(nn.Module):
    def __init__(self, vocabOcr: dict, input_size=(100, 300), lr0=1e-4, wandb_run=None):
        super().__init__()

        # --- Vocabs ---
        self.vocabOcr = vocabOcr
        self.blank_idx = self.vocabOcr["<BLANK>"]

        # --- Runtime / Backends ---
        self. sess = tf.Session(config=tf.ConfigProto(device_count={'GPU': 0}))
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # --- Backbone ---
        self.AryadCNN = AryadCNN()
        # Déterminer input_dim
        dim_in = torch.randn(1, 1, *input_size)
        with torch.no_grad():
            dim_out = self.AryadCNN(dim_in)
        input_dim = dim_out.shape[-1]

        self.context_encoder = ContextEncoder(input_dim=input_dim)
        self.decoder = DecoderCTC(d_model=256, num_classes=len(self.vocabOcr))

        # --- Optimizer / Loss ---
        self.optimizer = torch.optim.Adam(self.parameters(), lr=lr0)
        self.ctc_loss = nn.CTCLoss(blank=self.blank_idx, zero_infinity=True)

        # --- Logger ---
        self.wandb_run = wandb_run

    def forward(self, x):
        x = self.AryadCNN(x)                  # (B, T, C)
        x = self.context_encoder(x)           # (B, T, D)
        x = self.decoder(x)                   # (B, T, C)
        return x.permute(1, 0, 2)             # (T, B, C)
    
    def close(self):
        if self.sess is not None:
            self.sess.close()
            self.sess = None

    def fit(self, train_loader, val_loader=None, epochs=10, use_beam=True, beam_width=10, early_stopping_patience=5):
        self.to(self.device)
        best_val_loss = float('inf')
        epochs_no_improve = 0
        for epoch in range(1, epochs + 1):
            # ----------------- TRAIN -----------------
            self.train()
            train_loss = 0.0
            pbar = tqdm(train_loader, desc=f"[Epoch {epoch}/{epochs}] Training")
            for batch in pbar:
                images = batch['pixel_values'].to(self.device)
                labels = batch['labels'].to(self.device)
                label_lengths = batch['label_lengths'].to(self.device)

                logits = self(images)                 # (T, B, C)
                log_probs = nn.functional.log_softmax(logits, dim=2)
                input_lengths = torch.full(
                    (log_probs.size(1),),
                    log_probs.size(0),
                    dtype=torch.long
                ).to(self.device)
                targets = torch.cat([labels[i, :l] for i, l in enumerate(label_lengths)])

                loss = self.ctc_loss(log_probs, targets, input_lengths, label_lengths)

                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

                train_loss += loss.item()
                pbar.set_postfix(loss=loss.item())

                # --- Nettoyage mémoire ---
                del logits, log_probs, targets, input_lengths
                torch.cuda.empty_cache()

            avg_train_loss = train_loss / len(train_loader)

            # ----------------- VALIDATION -----------------
            avg_val_loss, avg_car, avg_war = None, None, None
            if val_loader is not None:
                self.eval()
                val_loss = 0.0
                with torch.no_grad():
                    for batch in tqdm(val_loader, desc=f"[Epoch {epoch}/{epochs}] Val"):
                        images = batch['pixel_values'].to(self.device)
                        labels = batch['labels'].to(self.device)
                        label_lengths = batch['label_lengths'].to(self.device)

                        logits = self(images)
                        log_probs = nn.functional.log_softmax(logits, dim=2)
                        input_lengths = torch.full(
                            (log_probs.size(1),),
                            log_probs.size(0),
                            dtype=torch.long
                        ).to(self.device)
                        targets = torch.cat([labels[i, :l] for i, l in enumerate(label_lengths)])
                        loss = self.ctc_loss(log_probs, targets, input_lengths, label_lengths)
                        val_loss += loss.item()

                        # --- Décodage et métriques ---
                        batch_predictions = self.ctc_beam_decode_torch_to_tf(log_probs,beam=use_beam, beam_width=beam_width)
                        batch_ground_truths = [
                            [list(self.vocabOcr)[idx] for idx in labels[i][:label_lengths[i]] if list(self.vocabOcr)[idx] not in ['<PAD>', '<BLANK>']]
                            for i in range(labels.size(0))
                        ]

                        preds = ["".join(word.split('|')) for word in batch_predictions]
                        gt = ["".join(ph) for ph in batch_ground_truths]

                        avg_car = general_car(gt,preds) if preds else 0.0
                        avg_war = general_war(gt,preds) if preds else 0.0

                        # --- Nettoyage mémoire ---
                        del logits, log_probs, targets, input_lengths
                        torch.cuda.empty_cache()

                    avg_val_loss = val_loss / max(1, len(val_loader))
                # ----------------- LOG -----------------
                msg = f"Epoch {epoch} | Train Loss: {avg_train_loss:.4f}"
                if val_loader is not None:
                    msg += f" | Val Loss: {avg_val_loss:.4f} | CAR: {avg_car:.4f} | WAR: {avg_war:.4f}"
                print(msg)
                print("GT:",gt,"PREDI:",preds)
                
            # --- Logging W&B ---
            if self.wandb_run and val_loader is not None:
                self.wandb_run.log({
                    "train_loss": avg_train_loss,
                    "val_loss": avg_val_loss,
                    "car": avg_car,
                    "war": avg_war,
                    "epoch": epoch
                })

            # ----------------- EARLY STOP -----------------
            if val_loader is not None:
                if avg_val_loss < best_val_loss:
                    best_val_loss = avg_val_loss
                    epochs_no_improve = 0
                    torch.save({
                        'epoch': epoch,
                        'model_state_dict': self.state_dict(),
                        'optimizer_state_dict': self.optimizer.state_dict(),
                        'val_loss': avg_val_loss
                    }, "best_model.pt")
                else:
                    epochs_no_improve += 1
                    if epochs_no_improve >= early_stopping_patience:
                        print(f"Early stopping triggered after {epoch} epochs.")
                        break
        # --- Après toutes les epochs ---
        self.close()


    def predict(self,image, beam=True, beam_width=10):
        if not isinstance(image, np.ndarray):
            image = np.array(image)
        image = np.expand_dims(np.expand_dims(image, axis=0), axis=0)
        image = torch.tensor(image, dtype=torch.float32).to(self.device)

        self.eval()
        with torch.no_grad():
            logits = self(image)
            log_probs = nn.functional.log_softmax(logits, dim=2)
            preds = self.ctc_beam_decode_torch_to_tf(log_probs,beam=beam, beam_width=beam_width)

            # Nettoyage mémoire
            del logits, log_probs
            torch.cuda.empty_cache()

        return preds

    def ctc_beam_decode_torch_to_tf(self, log_probs,beam=True, beam_width=10):
        preds_ = log_probs.cpu().detach().numpy().astype(np.float32)  # (T, B, C)
        input_lengths = [preds_.shape[0]] * preds_.shape[1]

        # Décodage
        if beam and beam_width > 1:
            decodes, _ = tf.nn.ctc_beam_search_decoder(
                inputs=preds_,
                sequence_length=input_lengths,
                beam_width=beam_width,
                merge_repeated=True
            )
        else:
            decodes, _ = tf.nn.ctc_greedy_decoder(
                inputs=preds_,
                sequence_length=input_lengths,
                merge_repeated=True
            )
         # Session locale
        dense_decoded = self. sess.run(tf.sparse.to_dense(decodes[0], default_value=-1))  # (B, max_len)
        decoded_strings = []
        for seq in dense_decoded:
            decoded_str = "".join(
                [list(self.vocabOcr)[idx] + "|" for idx in seq if idx != -1 and list(self.vocabOcr)[idx] not in ["<PAD>", "<BLANK>"]]
            )
            decoded_strings.append(decoded_str[:-1] if decoded_str else "")

        # Nettoyage
        del preds_, dense_decoded
        return decoded_strings




from datasets import Dataset,load_dataset
from datasets import Dataset, concatenate_datasets
from torch.nn.utils.rnn import pad_sequence
import torch
from datasets import concatenate_datasets

def load_set(name_set:str):
    if(name_set.lower() == "set_a" ):
        dataset_set_a = load_dataset("aryadacademie/Ocr_Arabic_Merged_data_set",split='Ocr_Arabic_Merged_data_set_a')
        return dataset_set_a
    elif(name_set.lower() == "set_b" ):
        dataset_set_b = load_dataset("aryadacademie/Ocr_Arabic_Merged_data_set",split='Ocr_Arabic_Merged_data_set_b')
        return dataset_set_b   
    elif(name_set.lower() == "set_c" ):
        dataset_set_c = load_dataset("aryadacademie/Ocr_Arabic_Merged_data_set",split='Ocr_Arabic_Merged_data_set_c')
        return dataset_set_c
    elif(name_set.lower() == "set_d" ):
        dataset_d = load_dataset("aryadacademie/Ocr_Arabic_Merged_data_set",split='Ocr_Arabic_Merged_data_set_d')
        return dataset_d
    else:
        return None

def dataset_merged(list_set:list):
    gthru = ["set_a","set_b","set_c","set_d"]
    valid_sets = True
    for set_name in list_set:
        if set_name.lower() not in gthru:
            valid_sets = False
            break
    if valid_sets:
        all_datasets = [load_set(set_name.lower()) for set_name in list_set]
    else:
        assert False, "Invalid set name(s) provided. Valid names are: 'set_a', 'set_b', 'set_c', 'set_d'."
    return concatenate_datasets(all_datasets)

def filter_by_patterns(dataset:Dataset,patterns:list):
    def is_valid(example):
        phon = example["phonemes"]
        return not any(bad in phon for bad in patterns)
    return dataset.filter(is_valid)

class AryadCollator:
    def __init__(self, vocabOcrArabic, pad_token="<PAD>"):
        self.vocabOcrArabic = vocabOcrArabic
        self.pad_token_id = vocabOcrArabic[pad_token]

    def __call__(self, batch):
        images = [torch.tensor(ex["array"], dtype=torch.float32).unsqueeze(0) for ex in batch]
        labels = [torch.tensor([self.vocabOcrArabic.get(t) for t in ex["phonemes"].split("|")], dtype=torch.long) for ex in batch]

        images = torch.stack(images)
        labels_padded = pad_sequence(labels, batch_first=True, padding_value=self.pad_token_id)

        return {
            "pixel_values": images,  # (B, 1, H, W)
            "labels": labels_padded, # (B, L)
            "label_lengths": torch.tensor([len(l) for l in labels])
        }

def loader_batch_data(dataset:Dataset,batch_size:int,vocabOcrArabic:set,train:bool=True):
    from torch.utils.data import DataLoader
    collator = AryadCollator(vocabOcrArabic=vocabOcrArabic)
    if train:
        return DataLoader(dataset, batch_size=batch_size, shuffle=True, collate_fn=collator)
    else:
        return DataLoader(dataset, batch_size=batch_size, shuffle=False, collate_fn=collator)

